# -*- coding: utf-8 -*-
#
#% $Id$
#
#
# Copyright (C) 2002-2011
# The MeqTree Foundation &
# ASTRON (Netherlands Foundation for Research in Astronomy)
# P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>,
# or write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

from PyQt4.Qt import *
import math
import Kittens.utils
pyfits = Kittens.utils.import_pyfits();
import os.path

from Kittens.widgets import BusyIndicator
from TigGUI.Widgets import FileSelector
from Tigger.Tools import Imaging

DEG = math.pi/180;

from astLib.astWCS import WCS

class RestoreImageDialog (QDialog):
  def __init__ (self,parent,modal=True,flags=Qt.WindowFlags()):
    QDialog.__init__(self,parent,flags);
    self.setModal(modal);
    self.setWindowTitle("Restore model into image");
    lo = QVBoxLayout(self);
    lo.setMargin(10);
    lo.setSpacing(5);
    # file selector
    self.wfile_in = FileSelector(self,label="Input FITS file:",dialog_label="Input FITS file",default_suffix="fits",file_types="FITS files (*.fits *.FITS)",file_mode=QFileDialog.ExistingFile);
    lo.addWidget(self.wfile_in);
    self.wfile_out = FileSelector(self,label="Output FITS file:",dialog_label="Output FITS file",default_suffix="fits",file_types="FITS files (*.fits *.FITS)",file_mode=QFileDialog.AnyFile);
    lo.addWidget(self.wfile_out);
    # beam size
    lo1 = QHBoxLayout();
    lo.addLayout(lo1);
    lo1.setContentsMargins(0,0,0,0);
    lo1.addWidget(QLabel("Restoring beam FWHM, major axis:",self));
    self.wbmaj = QLineEdit(self);
    lo1.addWidget(self.wbmaj);
    lo1.addWidget(QLabel("\"     minor axis:",self));
    self.wbmin = QLineEdit(self);
    lo1.addWidget(self.wbmin);
    lo1.addWidget(QLabel("\"     P.A.:",self));
    self.wbpa = QLineEdit(self);
    lo1.addWidget(self.wbpa);
    lo1.addWidget(QLabel(u"\u00B0",self));
    for w in self.wbmaj,self.wbmin,self.wbpa:
      w.setValidator(QDoubleValidator(self));
    lo1 = QHBoxLayout();
    lo.addLayout(lo1);
    lo1.setContentsMargins(0,0,0,0);
    self.wfile_psf = FileSelector(self,label="Set restoring beam by fitting PSF image:",dialog_label="PSF FITS file",default_suffix="fits",file_types="FITS files (*.fits *.FITS)",file_mode=QFileDialog.ExistingFile);
    lo1.addSpacing(32);
    lo1.addWidget(self.wfile_psf);
    # selection only
    self.wselonly = QCheckBox("restore selected model sources only",self);
    lo.addWidget(self.wselonly );
    # OK/cancel buttons
    lo.addSpacing(10);
    lo2 = QHBoxLayout();
    lo.addLayout(lo2);
    lo2.setContentsMargins(0,0,0,0);
    lo2.setMargin(5);
    self.wokbtn = QPushButton("OK",self);
    self.wokbtn.setMinimumWidth(128);
    QObject.connect(self.wokbtn,SIGNAL("clicked()"),self.accept);
    self.wokbtn.setEnabled(False);
    cancelbtn = QPushButton("Cancel",self);
    cancelbtn.setMinimumWidth(128);
    QObject.connect(cancelbtn,SIGNAL("clicked()"),self.reject);
    lo2.addWidget(self.wokbtn);
    lo2.addStretch(1);
    lo2.addWidget(cancelbtn);
    self.setMinimumWidth(384);
    # signals
    QObject.connect(self.wfile_in,SIGNAL("filenameSelected"),self._fileSelected);
    QObject.connect(self.wfile_in,SIGNAL("filenameSelected"),self._inputFileSelected);
    QObject.connect(self.wfile_out,SIGNAL("filenameSelected"),self._fileSelected);
    QObject.connect(self.wfile_psf,SIGNAL("filenameSelected"),self._psfFileSelected);
    # internal state
    self.qerrmsg = QErrorMessage(self);

  def setModel (self,model):
    nsel = len([ src for src in model.sources if src.selected ]);
    self.wselonly.setVisible(nsel>0 and nsel<len(model.sources));
    self.model = model;
    self._fileSelected(None);

  def _fileSelected (self,filename):
    self.wokbtn.setEnabled(bool(self.wfile_in.filename() and self.wfile_out.filename()));
    
  def _inputFileSelected (self,filename):
    if filename:
      try:
        header = pyfits.open(filename)[0].header;
      except:
        self.qerrmsg.showMessage("Error reading FITS file %s: %s"%(filename,str(err)));
        self.wfile_in.setFilename("");
        return;
      # try to get beam extents
      gx,gy,grot = [ header.get(x,None) for x in 'BMAJ','BMIN','BPA' ];
      if all([x is not None for x in gx,gy,grot]):
        # if beam size is already set, ask before overwriting
        print [ str(x.text()) for x in self.wbmaj,self.wbmin,self.wbpa ]
        if any([ bool(str(x.text())) for x in self.wbmaj,self.wbmin,self.wbpa ]) and \
          QMessageBox.question(self,"Set restoring beam","Also reset restoring beam size from this FITS file?",
            QMessageBox.Yes|QMessageBox.No) != QMessageBox.Yes:
          return;
        self.wbmaj.setText("%.2f"%(gx*3600));
        self.wbmin.setText("%.2f"%(gy*3600));
        self.wbpa.setText("%.2f"%grot);
    
  def _psfFileSelected (self,filename):
    busy = BusyIndicator();
    filename = str(filename);
    self.parent().showMessage("Fitting gaussian to PSF file %s"%filename);
    try:
      bmaj,bmin,pa = [ x/DEG for x in Imaging.fitPsf(filename) ];
    except Exception,err:
      busy = None;
      self.qerrmsg.showMessage("Error fitting PSF file %s: %s"%(filename,str(err)));
      return;
    bmaj *= 3600*Imaging.FWHM;
    bmin *= 3600*Imaging.FWHM;
    self.wbmaj.setText(str(bmaj));
    self.wbmin.setText(str(bmin));
    self.wbpa.setText(str(pa));

  def accept (self):
    """Tries to restore the image, and closes the dialog if successful.""";
    # get list of sources to restore
    sources = self.model.sources;
    sel_sources = filter(lambda src:src.selected,sources);
    if len(sel_sources) > 0 and len(sel_sources) < len(sources) and self.wselonly.isChecked():
      sources = sel_sources;
    if not sources:
      self.qerrmsg.showMessage("No sources to restore.");
      return;
    busy = BusyIndicator();
    # get filenames
    infile = self.wfile_in.filename();
    outfile = self.wfile_out.filename();
    self.parent().showMessage("Restoring %d model sources to image %s, writing to %s"%(len(sources),infile,outfile));
    # read fits file
    try:
      input_hdu = pyfits.open(infile)[0];
    except Exception,err:
      busy = None;
      self.qerrmsg.showMessage("Error reading FITS file %s: %s"%(infile,str(err)));
      return;
    # get beam sizes
    try:
      bmaj = float(str(self.wbmaj.text()));
      bmin = float(str(self.wbmin.text()));
      pa = float(str(self.wbpa.text()) or "0");
    except Exception,err:
      busy = None;
      self.qerrmsg.showMessage("Invalid beam size specified");
      return;
    bmaj = bmaj/(Imaging.FWHM*3600)*DEG;
    bmin = bmin/(Imaging.FWHM*3600)*DEG;
    pa = pa*DEG;
    # restore
    try:
      Imaging.restoreSources(input_hdu,sources,bmaj,bmin,pa);
    except Exception,err:
      busy = None;
      self.qerrmsg.showMessage("Error restoring model into image: %s"%str(err));
      return;
    # save fits file
    try:
      input_hdu.writeto(outfile,clobber=True);
    except Exception,err:
      busy = None;
      self.qerrmsg.showMessage("Error writing FITS file %s: %s"%(outfile,str(err)));
      return;
    self.parent().loadImage(outfile);
    busy = None;
    return QDialog.accept(self);

def restore_into_image (mainwin,model):
  dialog = getattr(mainwin,'_restore_into_image_dialog',None);
  if not dialog:
    dialog = mainwin._restore_into_image_dialog = RestoreImageDialog(mainwin);
  dialog.setModel(model);
  # show dialog
  return dialog.exec_();

from TigGUI.Tools import registerTool
registerTool("Restore model into image...",restore_into_image);



# -*- coding: utf-8 -*-

#
#% $Id$
#
#
# Copyright (C) 2002-2011
# The MeqTree Foundation &
# ASTRON (Netherlands Foundation for Research in Astronomy)
# P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>,
# or write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

from TigGUI import *

import os
import os.path
import time
import sys
import fnmatch
import traceback

from PyQt4.Qt import *

import Kittens.utils
from Kittens.utils import PersistentCurrier

from Tigger.Models import ModelClasses
from Tigger.Models import SkyModel
from Tigger.Models.Formats import ModelHTML
import Widgets
import AboutDialog
from SkyModelTreeWidget import *
from Plot.SkyModelPlot import *
from Images.Manager import ImageManager
import TigGUI.Tools.source_selector

_verbosity = Kittens.utils.verbosity(name="mainwin");
dprint = _verbosity.dprint;
dprintf = _verbosity.dprintf;

class MainWindow (QMainWindow):
  ViewModelColumns = [ "name","RA","Dec","type","Iapp","I","Q","U","V","RM","spi","shape" ];
  def __init__ (self,parent,hide_on_close=False):
    QMainWindow.__init__(self,parent);
    self.setWindowIcon(pixmaps.tigger_starface.icon());
    self._currier = PersistentCurrier();
    self.hide();
    # init column constants
    for icol,col in enumerate(self.ViewModelColumns):
        setattr(self,"Column%s"%col.capitalize(),icol);
    # init GUI
    self.setWindowTitle("Tigger");
    # self.setIcon(pixmaps.purr_logo.pm());
    cw = QWidget(self);
    self.setCentralWidget(cw);
    cwlo = QVBoxLayout(cw);
    cwlo.setMargin(5);
    # make splitter
    spl1 = self._splitter1 = QSplitter(Qt.Vertical,cw);
    spl1.setOpaqueResize(False);
    cwlo.addWidget(spl1);
    # Create listview of LSM entries
    self.tw = SkyModelTreeWidget(spl1);
    self.tw.hide();

    # split bottom pane
    spl2 = self._splitter2 = QSplitter(Qt.Horizontal,spl1);
    spl2.setOpaqueResize(False);
    self._skyplot_stack = QWidget(spl2);
    self._skyplot_stack_lo = QVBoxLayout(self._skyplot_stack);
    self._skyplot_stack_lo.setContentsMargins(0,0,0,0);

    # add plot
    self.skyplot = SkyModelPlotter(self._skyplot_stack,self);
    self.skyplot.resize(128,128);
    self.skyplot.setSizePolicy(QSizePolicy.Minimum,QSizePolicy.Preferred);
    self._skyplot_stack_lo.addWidget(self.skyplot,1000);
    self.skyplot.hide();
    QObject.connect(self.skyplot,SIGNAL("imagesChanged"),self._imagesChanged);
    QObject.connect(self.skyplot,SIGNAL("showMessage"),self.showMessage);
    QObject.connect(self.skyplot,SIGNAL("showErrorMessage"),self.showErrorMessage);

    self._grouptab_stack = QWidget(spl2);
    self._grouptab_stack_lo = lo =QVBoxLayout(self._grouptab_stack);
    self._grouptab_stack_lo.setContentsMargins(0,0,0,0);
    # add groupings table
    self.grouptab = ModelGroupsTable(self._grouptab_stack);
    self.grouptab.setSizePolicy(QSizePolicy.Preferred,QSizePolicy.Preferred);
    QObject.connect(self,SIGNAL("hasSkyModel"),self.grouptab.setEnabled);
    lo.addWidget(self.grouptab,1000);
    lo.addStretch(1);
    self.grouptab.hide();

    # add image controls -- parentless for now (setLayout will reparent them anyway)
    self.imgman = ImageManager();
    self.skyplot.setImageManager(self.imgman);
    QObject.connect(self.imgman,SIGNAL("imagesChanged"),self._imagesChanged);
    QObject.connect(self.imgman,SIGNAL("showMessage"),self.showMessage);
    QObject.connect(self.imgman,SIGNAL("showErrorMessage"),self.showErrorMessage);

    # enable status line
    self.statusBar().show();
    # Create and populate main menu
    menubar = self.menuBar();
    # File menu
    file_menu = menubar.addMenu("&File");
    qa_open = file_menu.addAction("&Open model...",self._openFileCallback,Qt.CTRL+Qt.Key_O);
    qa_merge = file_menu.addAction("&Merge in model...",self._mergeFileCallback,Qt.CTRL+Qt.SHIFT+Qt.Key_O);
    QObject.connect(self,SIGNAL("hasSkyModel"),qa_merge.setEnabled);
    file_menu.addSeparator();
    qa_save = file_menu.addAction("&Save model",self.saveFile,Qt.CTRL+Qt.Key_S);
    QObject.connect(self,SIGNAL("isUpdated"),qa_save.setEnabled);
    qa_save_as = file_menu.addAction("Save model &as...",self.saveFileAs);
    QObject.connect(self,SIGNAL("hasSkyModel"),qa_save_as.setEnabled);
    qa_save_selection_as = file_menu.addAction("Save selection as...",self.saveSelectionAs);
    QObject.connect(self,SIGNAL("hasSelection"),qa_save_selection_as.setEnabled);
    file_menu.addSeparator();
    qa_close = file_menu.addAction("&Close model",self.closeFile,Qt.CTRL+Qt.Key_W);
    QObject.connect(self,SIGNAL("hasSkyModel"),qa_close.setEnabled);
    qa_quit = file_menu.addAction("Quit",self.close,Qt.CTRL+Qt.Key_Q);

    # Image menu
    menubar.addMenu(self.imgman.getMenu());
    # Plot menu
    menubar.addMenu(self.skyplot.getMenu());

    # LSM Menu
    em = QMenu("&LSM",self);
    self._qa_em = menubar.addMenu(em);
    self._qa_em.setVisible(False);
    QObject.connect(self,SIGNAL("hasSkyModel"),self._qa_em.setVisible);
    self._column_view_menu = QMenu("&Show columns",self);
    self._qa_cv_menu = em.addMenu(self._column_view_menu);
    em.addSeparator();
    em.addAction("Select &all",self._selectAll,Qt.CTRL+Qt.Key_A);
    em.addAction("&Invert selection",self._selectInvert,Qt.CTRL+Qt.Key_I);
    em.addAction("Select b&y attribute...",self._showSourceSelector,Qt.CTRL+Qt.Key_Y);
    em.addSeparator();
    qa_add_tag = em.addAction("&Tag selection...",self.addTagToSelection,Qt.CTRL+Qt.Key_T);
    QObject.connect(self,SIGNAL("hasSelection"),qa_add_tag.setEnabled);
    qa_del_tag = em.addAction("&Untag selection...",self.removeTagsFromSelection,Qt.CTRL+Qt.Key_U);
    QObject.connect(self,SIGNAL("hasSelection"),qa_del_tag.setEnabled);
    qa_del_sel = em.addAction("&Delete selection",self._deleteSelection);
    QObject.connect(self,SIGNAL("hasSelection"),qa_del_sel.setEnabled);

   # Tools menu
    tm = self._tools_menu = QMenu("&Tools",self);
    self._qa_tm = menubar.addMenu(tm);
    self._qa_tm.setVisible(False);
    QObject.connect(self,SIGNAL("hasSkyModel"),self._qa_tm.setVisible);

   # Help menu
    menubar.addSeparator();
    hm = self._help_menu = menubar.addMenu("&Help");
    hm.addAction("&About...",self._showAboutDialog);
    self._about_dialog = None;

    # message handlers
    self.qerrmsg = QErrorMessage(self);

    # set initial state
    self.setAcceptDrops(True);
    self.model = None;
    self.filename = None;
    self._display_filename = None;
    self._open_file_dialog = self._merge_file_dialog = self._save_as_dialog = self._save_sel_as_dialog = self._open_image_dialog = None;
    self.emit(SIGNAL("isUpdated"),False);
    self.emit(SIGNAL("hasSkyModel"),False);
    self.emit(SIGNAL("hasSelection"),False);
    self._exiting = False;

    # set initial layout
    self._current_layout = None;
    self.setLayout(self.LayoutEmpty);
    dprint(1,"init complete");

  # layout identifiers
  LayoutEmpty = "empty";
  LayoutImage = "image";
  LayoutImageModel = "model";
  LayoutSplit = "split";

  def _getFilenamesFromDropEvent (self,event):
    """Checks if drop event is valid (i.e. contains a local URL to a FITS file), and returns list of filenames contained therein.""";
    dprint(1,"drop event:",event.mimeData().text());
    if not event.mimeData().hasUrls():
      dprint(1,"drop event: no urls");
      return None;
    filenames = [];
    for url in event.mimeData().urls():
      name = str(url.toLocalFile());
      dprint(2,"drop event: name is",name);
      if name and Images.isFITS(name):
        filenames.append(name);
    dprint(2,"drop event: filenames are",filenames);
    return filenames;

  def dragEnterEvent (self,event):
    if self._getFilenamesFromDropEvent(event):
      dprint(1,"drag-enter accepted");
      event.acceptProposedAction();
    else:
      dprint(1,"drag-enter rejected");

  def dropEvent (self,event):
    filenames = self._getFilenamesFromDropEvent(event);
    dprint(1,"dropping",filenames);
    if filenames:
      event.acceptProposedAction();
      busy = BusyIndicator();
      for name in filenames:
        self.imgman.loadImage(name);

  def saveSizes (self):
    if self._current_layout is not None:
      dprint(1,"saving sizes for layout",self._current_layout);
      # save main window size and splitter dimensions
      sz = self.size();
      Config.set('%s-main-window-width'%self._current_layout,sz.width());
      Config.set('%s-main-window-height'%self._current_layout,sz.height());
      for spl,name in ((self._splitter1,"splitter1"),(self._splitter2,"splitter2")):
        ssz = spl.sizes();
        for i,sz in enumerate(ssz):
          Config.set('%s-%s-size%d'%(self._current_layout,name,i),sz);

  def loadSizes (self):
    if self._current_layout is not None:
      dprint(1,"loading sizes for layout",self._current_layout);
      # get main window size and splitter dimensions
      w = Config.getint('%s-main-window-width'%self._current_layout,0);
      h = Config.getint('%s-main-window-height'%self._current_layout,0);
      dprint(2,"window size is",w,h);
      if not (w and h):
        return None;
      self.resize(QSize(w,h));
      for spl,name in (self._splitter1,"splitter1"),(self._splitter2,"splitter2"):
        ssz = [ Config.getint('%s-%s-size%d'%(self._current_layout,name,i),-1) for i in 0,1 ];
        dprint(2,"splitter",name,"sizes",ssz);
        if all([ sz >=0 for sz in ssz ]):
          spl.setSizes(ssz);
        else:
          return None;
    return True;

  def setLayout (self,layout):
    """Changes the current window layout. Restores sizes etc. from config file.""";
    if self._current_layout is layout:
      return;
    dprint(1,"switching to layout",layout);
    # save sizes to config file
    self.saveSizes();
    # remove imgman widget from all layouts
    for lo in self._skyplot_stack_lo,self._grouptab_stack_lo:
      if lo.indexOf(self.imgman) >= 0:
        lo.removeWidget(self.imgman);
    # assign it to appropriate parent and parent's layout
    if layout is self.LayoutImage or layout is self.LayoutEmpty:
      lo = self._skyplot_stack_lo;
    else:
      lo = self._grouptab_stack_lo;
    self.imgman.setParent(lo.parentWidget());
    lo.addWidget(self.imgman,0);
    # show/hide panels
    if layout is self.LayoutEmpty:
      self.tw.hide();
      self.grouptab.hide();
      self.skyplot.show();
    elif layout is self.LayoutImage:
      self.tw.hide();
      self.grouptab.hide();
      self.skyplot.show();
    elif layout is self.LayoutImageModel:
      self.tw.show();
      self.grouptab.show();
      self.skyplot.show();
    # reload sizes
    self._current_layout = layout;
    if not self.loadSizes():
      dprint(1,"no sizes loaded, setting defaults");
      if layout is self.LayoutEmpty:
        self.resize(QSize(512,256));
      elif layout is self.LayoutImage:
        self.resize(QSize(512,512));
        self._splitter2.setSizes([512,0]);
      elif layout is self.LayoutImageModel:
        self.resize(QSize(1024,512));
        self._splitter1.setSizes([256,256]);
        self._splitter2.setSizes([256,256]);

  def enableUpdates (self,enable=True):
    """Enables updates of the child widgets. Usually called after startup is completed (i.e. all data loaded)""";
    self.skyplot.enableUpdates(enable);
    if enable:
      if self.model:
        self.setLayout(self.LayoutImageModel);
      elif self.imgman.getImages():
        self.setLayout(self.LayoutImage);
      else:
        self.setLayout(self.LayoutEmpty);
      self.show();

  def _showAboutDialog (self):
    if not self._about_dialog:
      self._about_dialog = AboutDialog.AboutDialog(self);
    self._about_dialog.show();

  def addTool (self,name,callback):
    """Adds a tool to the Tools menu""";
    self._tools_menu.addAction(name,self._currier.curry(self._callTool,callback));

  def _callTool (self,callback):
    callback(self,self.model);

  def _imagesChanged (self):
    """Called when the set of loaded images has changed""";
    if self.imgman.getImages():
      if self._current_layout  is self.LayoutEmpty:
       self.setLayout(self.LayoutImage);
    else:
      if not self.model:
       self.setLayout(self.LayoutEmpty);

  def _selectAll (self):
    if not self.model:
      return;
    busy = BusyIndicator();
    for src in self.model.sources:
      src.selected = True;
    self.model.emitSelection(self);

  def _selectInvert (self):
    if not self.model:
      return;
    busy = BusyIndicator();
    for src in self.model.sources:
      src.selected = not src.selected;
    self.model.emitSelection(self);

  def _deleteSelection (self):
    unselected = [ src for src in self.model.sources if not src.selected ];
    nsel = len(self.model.sources) - len(unselected);
    if QMessageBox.question(self,"Delete selection","""<P>Really deleted %d selected source(s)?
        %d unselected sources will remain in the model.</P>"""%(nsel,len(unselected)),
        QMessageBox.Ok|QMessageBox.Cancel,QMessageBox.Cancel) != QMessageBox.Ok:
      return;
    self.model.setSources(unselected);
    self.showMessage("""Deleted %d sources"""%nsel);
    self.model.emitUpdate(SkyModel.UpdateAll,origin=self);

  def _showSourceSelector (self):
    TigGUI.Tools.source_selector.show_source_selector(self, self.model);

  def _updateModelSelection (self,num,origin=None):
    """Called when the model selection has been updated.""";
    self.emit(SIGNAL("hasSelection"),bool(num));

  import Tigger.Models.Formats
  _formats = [f[1] for f in Tigger.Models.Formats.listFormatsFull()];

  _load_file_types = [ (doc,["*"+ext for ext in extensions],load) for load,save,doc,extensions in _formats if load ];
  _save_file_types = [ (doc,["*"+ext for ext in extensions],save) for load,save,doc,extensions in _formats if save ];

  def showMessage (self,msg,time=3000):
    self.statusBar().showMessage(msg,3000);

  def showErrorMessage (self,msg,time=3000):
    self.qerrmsg.showMessage(msg);

  def loadImage (self,filename):
    return self.imgman.loadImage(filename);

  def setModel (self,model):
    self.emit(SIGNAL("modelChanged"),model);
    if model:
      self.model = model;
      self.emit(SIGNAL("hasSkyModel"),True);
      self.emit(SIGNAL("hasSelection"),False);
      self.emit(SIGNAL("isUpdated"),False);
      self.model.enableSignals();
      self.model.connect("updated",self._indicateModelUpdated);
      self.model.connect("selected",self._updateModelSelection);
      # pass to children
      self.tw.setModel(self.model);
      self.grouptab.setModel(self.model);
      self.skyplot.setModel(self.model);
      # add items to View menu
      self._column_view_menu.clear();
      self.tw.addColumnViewActionsTo(self._column_view_menu);
    else:
      self.model = None;
      self.setWindowTitle("Tigger");
      self.emit(SIGNAL("hasSelection"),False);
      self.emit(SIGNAL("isUpdated"),False);
      self.emit(SIGNAL("hasSkyModel"),False);
      self.tw.clear();
      self.grouptab.clear();
      self.skyplot.setModel(None);

  def _openFileCallback (self):
    if not self._open_file_dialog:
      filters = ";;".join([  "%s (%s)"%(name," ".join(patterns)) for name,patterns,func in self._load_file_types ]);
      dialog = self._open_file_dialog = QFileDialog(self,"Open sky model",".",filters);
      dialog.setFileMode(QFileDialog.ExistingFile);
      dialog.setModal(True);
      QObject.connect(dialog,SIGNAL("filesSelected(const QStringList &)"),self.openFile);
    self._open_file_dialog.exec_();
    return;

  def _mergeFileCallback (self):
    if not self._merge_file_dialog:
      filters = ";;".join([  "%s (%s)"%(name," ".join(patterns)) for name,patterns,func in self._load_file_types ]);
      dialog = self._merge_file_dialog = QFileDialog(self,"Merge in sky model",".",filters);
      dialog.setFileMode(QFileDialog.ExistingFile);
      dialog.setModal(True);
      QObject.connect(dialog,SIGNAL("filesSelected(const QStringList &)"),
        self._currier.curry(self.openFile,merge=True));
    self._merge_file_dialog.exec_();
    return;

  def openFile (self,filename=None,format=None,merge=False,show=True):
    # check that we can close existing model
    if not merge and not self._canCloseExistingModel():
      return False;
    if isinstance(filename,QStringList):
      filename = filename[0];
    filename = str(filename);
    # try to determine the file type
    filetype,import_func,export_func,doc = Tigger.Models.Formats.resolveFormat(filename, format);
    if import_func is None:
      self.showErrorMessage("""Error loading model file %s: unknown file format"""%filename);
      return;
    # try to load the specified file
    busy = BusyIndicator();
    self.showMessage("""Reading %s file %s"""%(filetype,filename),3000);
    QApplication.flush();
    try:
      model = import_func(filename);
      model.setFilename(filename);
    except:
      busy = None;
      self.showErrorMessage("""Error loading '%s' file %s: %s"""%(filetype,filename,str(sys.exc_info()[1])));
      return;
    # set the layout
    if show:
      self.setLayout(self.LayoutImageModel);
    # add to content
    if merge and self.model:
      self.model.addSources(model.sources);
      self.showMessage("""Merged in %d sources from '%s' file %s"""%(len(model.sources),filetype,filename),3000);
      self.model.emitUpdate(SkyModel.UpdateAll);
    else:
      self.showMessage("""Loaded %d sources from '%s' file %s"""%(len(model.sources),filetype,filename),3000);
      self._display_filename = os.path.basename(filename);
      self.setModel(model);
      self._indicateModelUpdated(updated=False);
      # only set self.filename if an export function is available for this format. Otherwise set it to None, so that trying to save
      # the file results in a save-as operation (so that we don't save to a file in an unsupported format).
      self.filename = filename if export_func else None;

  def closeEvent (self,event):
    dprint(1,"closing");
    self._exiting = True;
    self.saveSizes();
    if not self.closeFile():
      self._exiting = False;
      event.ignore();
      return;
    self.skyplot.close();
    self.imgman.close();
    self.emit(SIGNAL("closing"));
    dprint(1,"invoking os._exit(0)");
    os._exit(0);
    QMainWindow.closeEvent(self,event);

  def _canCloseExistingModel (self):
    # save model if modified
    if self.model and self._model_updated:
      res = QMessageBox.question(self,"Closing sky model","<P>Model has been modified, would you like to save the changes?</P>",
                    QMessageBox.Save|QMessageBox.Discard|QMessageBox.Cancel,QMessageBox.Save);
      if res == QMessageBox.Cancel:
        return False;
      elif res == QMessageBox.Save:
        if not self.saveFile(confirm=False,overwrite=True):
          return False;
    # unload model images, unless we are already exiting anyway
    if not self._exiting:
      self.imgman.unloadModelImages();
    return True;

  def closeFile (self):
    if not self._canCloseExistingModel():
      return False;
    # close model
    self._display_filename = None;
    self.setModel(None);
    # set the layout
    self.setLayout(self.LayoutImage if self.imgman.getTopImage() else self.LayoutEmpty);
    return True;

  def saveFile (self,filename=None,confirm=False,overwrite=True,non_native=False):
    """Saves file using the specified 'filename'. If filename is None, uses current filename, if
    that is not set, goes to saveFileAs() to open dialog and get a filename.
    If overwrite=False, will ask for confirmation before overwriting an existing file.
    If non_native=False, will ask for confirmation before exporting in non-native format.
    If confirm=True, will ask for confirmation regardless.
    Returns True if saving succeeded, False on error (or if cancelled by user).
    """;
    if isinstance(filename,QStringList):
      filename = filename[0];
    filename = ( filename and str(filename) ) or self.filename;
    if filename is None:
      return self.saveFileAs();
    else:
      warning = '';
      # try to determine the file type
      filetype,import_func,export_func,doc = Tigger.Models.Formats.resolveFormat(filename, None);
      if export_func is None:
        self.showErrorMessage("""Error saving model file %s: unsupported output format"""%filename);
        return;
      if os.path.exists(filename) and not overwrite:
        warning += "<P>The file already exists and will be overwritten.</P>";
      if filetype != 'Tigger' and not non_native:
        warning += """<P>Please note that you are exporting the model using the external format '%s'.
              Source types, tags and other model features not supported by this
              format will be omitted during the export.</P>"""%filetype;
      # get confirmation
      if confirm or warning:
        dialog = QMessageBox.warning if warning else QMessageBox.question;
        if dialog(self,"Saving sky model","<P>Save model to %s?</P>%s"%(filename,warning),
                  QMessageBox.Save|QMessageBox.Cancel,QMessageBox.Save) != QMessageBox.Save:
          return False;
      busy = BusyIndicator();
      try:
        export_func(self.model,filename);
        self.model.setFilename(filename);
      except:
          busy = None;
          self.showErrorMessage("""Error saving model file %s: %s"""%(filename,str(sys.exc_info()[1])));
          return False;
      self.showMessage("""Saved model to file %s"""%filename,3000);
      self._display_filename = os.path.basename(filename);
      self._indicateModelUpdated(updated=False);
      self.filename = filename;
      return True;

  def saveFileAs (self,filename=None):
    """Saves file using the specified 'filename'. If filename is None, opens dialog to get a filename.
    Returns True if saving succeeded, False on error (or if cancelled by user).
    """;
    if filename is None:
      if not self._save_as_dialog:
          filters = ";;".join([  "%s (%s)"%(name," ".join(patterns)) for name,patterns,func in self._save_file_types ]);
          dialog = self._save_as_dialog = QFileDialog(self,"Save sky model",".",filters);
          dialog.setDefaultSuffix(ModelHTML.DefaultExtension);
          dialog.setFileMode(QFileDialog.AnyFile);
          dialog.setAcceptMode(QFileDialog.AcceptSave);
          dialog.setConfirmOverwrite(False);
          dialog.setModal(True);
          QObject.connect(dialog,SIGNAL("filesSelected(const QStringList &)"),self.saveFileAs);
      return self._save_as_dialog.exec_() == QDialog.Accepted;
    # filename supplied, so save
    return self.saveFile(filename,confirm=False);

  def saveSelectionAs (self,filename=None,force=False):
    if not self.model:
      return;
    if filename is None:
      if not self._save_sel_as_dialog:
          filters = ";;".join([  "%s (%s)"%(name," ".join(patterns)) for name,patterns,func in self._save_file_types ]);
          dialog = self._save_sel_as_dialog = QFileDialog(self,"Save sky model",".",filters);
          dialog.setDefaultSuffix(ModelHTML.DefaultExtension);
          dialog.setFileMode(QFileDialog.AnyFile);
          dialog.setAcceptMode(QFileDialog.AcceptSave);
          dialog.setConfirmOverwrite(True);
          dialog.setModal(True);
          QObject.connect(dialog,SIGNAL("filesSelected(const QStringList &)"),self.saveSelectionAs);
      return self._save_sel_as_dialog.exec_() == QDialog.Accepted;
    # save selection
    if isinstance(filename,QStringList):
      filename = filename[0];
    filename= str(filename);
    selmodel = self.model.copy();
    sources = [ src for src in self.model.sources if src.selected ];
    if not sources:
      self.showErrorMessage("""You have not selected any sources to save.""");
      return;
    # try to determine the file type
    filetype,import_func,export_func,doc = Tigger.Models.Formats.resolveFormat(filename, None);
    if export_func is None:
      self.showErrorMessage("""Error saving model file %s: unsupported output format"""%filename);
      return;
    busy = BusyIndicator();
    try:
      export_func(self.model,filename,sources=sources);
    except:
      busy = None;
      self.showErrorMessage("""Error saving selection to model file %s: %s"""%(filename,str(sys.exc_info()[1])));
      return False;
    self.showMessage("""Wrote %d selected source%s to file %s"""%(len(selmodel.sources),"" if len(selmodel.sources)==1 else "s",filename),3000);
    pass;

  def addTagToSelection (self):
    if not hasattr(self,'_add_tag_dialog'):
      self._add_tag_dialog = Widgets.AddTagDialog(self,modal=True);
    self._add_tag_dialog.setTags(self.model.tagnames);
    self._add_tag_dialog.setValue(True);
    if self._add_tag_dialog.exec_() != QDialog.Accepted:
      return;
    tagname,value = self._add_tag_dialog.getTag();
    if tagname is None or value is None:
      return None;
    dprint(1,"tagging selected sources with",tagname,value);
    # tag selected sources
    for src in self.model.sources:
      if src.selected:
        src.setAttribute(tagname,value);
    # If tag is not new, set a UpdateSelectionOnly flag on the signal
    dprint(1,"adding tag to model");
    self.model.addTag(tagname);
    dprint(1,"recomputing totals");
    self.model.getTagGrouping(tagname).computeTotal(self.model.sources);
    dprint(1,"emitting update signal");
    what = SkyModel.UpdateSourceContent+SkyModel.UpdateTags+SkyModel.UpdateSelectionOnly;
    self.model.emitUpdate(what,origin=self);

  def removeTagsFromSelection (self):
    if not hasattr(self,'_remove_tag_dialog'):
      self._remove_tag_dialog = Widgets.SelectTagsDialog(self,modal=True,caption="Remove Tags",ok_button="Remove");
    # get set of all tags in selected sources
    tags = set();
    for src in self.model.sources:
      if src.selected:
        tags.update(src.getTagNames());
    if not tags:
      return;
    tags = list(tags);
    tags.sort();
    # show dialog
    self._remove_tag_dialog.setTags(tags);
    if self._remove_tag_dialog.exec_() != QDialog.Accepted:
      return;
    tags = self._remove_tag_dialog.getSelectedTags();
    if not tags:
      return;
    # ask for confirmation
    plural = (len(tags)>1 and "s") or "";
    if QMessageBox.question(self,"Removing tags","<P>Really remove the tag%s '%s' from selected sources?</P>"%(plural,"', '".join(tags)),
            QMessageBox.Yes|QMessageBox.No,QMessageBox.Yes) != QMessageBox.Yes:
      return;
    # remove the tags
    for src in self.model.sources:
      if src.selected:
        for tag in tags:
          src.removeAttribute(tag);
    # update model
    self.model.scanTags();
    self.model.initGroupings();
    # emit signal
    what = SkyModel.UpdateSourceContent+SkyModel.UpdateTags+SkyModel.UpdateSelectionOnly;
    self.model.emitUpdate(what,origin=self);

  def _indicateModelUpdated (self,what=None,origin=None,updated=True):
    """Marks model as updated.""";
    self._model_updated = updated;
    self.emit(SIGNAL("isUpdated"),updated);
    if self.model:
      self.setWindowTitle("Tigger - %s%s"%((self._display_filename or "(unnamed)"," (modified)" if updated else "")));




# -*- coding: utf-8 -*-
#
#% $Id$
#
#
# Copyright (C) 2002-2011
# The MeqTree Foundation &
# ASTRON (Netherlands Foundation for Research in Astronomy)
# P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>,
# or write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

import sys

import Kittens.config
import os.path
import Tigger
from Tigger import import_pyfits, nuke_matplotlib


__version__ = "1.4.0"

release_string = __version__
svn_revision_string = __version__
svn_revision_html = __version__

# initializes GUI-related globals. Only called from the viewer
def init_gui():
    from Kittens.widgets import BusyIndicator
    import Kittens.pixmaps
    import Kittens.utils
    global pixmaps, Config, ConfigFile, ConfigFileName
    pixmaps = Kittens.pixmaps.PixmapCache("TigGUI")
    ConfigFileName = ".tigger.conf"
    ConfigFile = Kittens.config.DualConfigParser("tigger.conf",["/usr/lib/TigGUI", os.path.dirname(__file__)])
    Config = Kittens.config.SectionParser(ConfigFile,"Tigger")


startup_dprint = startup_dprintf = lambda *dum:None
_verbosity = Kittens.utils.verbosity(name="tiggui")
dprint = _verbosity.dprint
dprintf = _verbosity.dprintf





# -*- coding: utf-8 -*-
#
#% $Id$ 
#
#
# Copyright (C) 2002-2011
# The MeqTree Foundation & 
# ASTRON (Netherlands Foundation for Research in Astronomy)
# P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>,
# or write to the Free Software Foundation, Inc., 
# 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

import sys
import math
import traceback
import re

from PyQt4.Qt import *
from PyQt4.Qwt5 import *

class TiggerPlotCurve (QwtPlotCurve):
  """Wrapper around QwtPlotCurve to make it compatible with numpy float types"""
  def setData (self,x,y):
    return QwtPlotCurve.setData(self,map(float,x),map(float,y));

class TiggerPlotMarker (QwtPlotMarker):
  """Wrapper around QwtPlotCurve to make it compatible with numpy float types"""
  def setValue (self,x,y):
    return QwtPlotMarker.setValue(self,float(x),float(y));


class FloatValidator (QValidator):
  """QLineEdit validator for float items in standard or scientific notation""";
  re_intermediate = re.compile("^-?([0-9]*)\.?([0-9]*)([eE]([+-])?[0-9]*)?$");
  def validate (self,input,pos):
    input = str(input);
    try:
      x = float(input);
      return QValidator.Acceptable,pos;
    except:
      pass;
    if not input or self.re_intermediate.match(input):
      return QValidator.Intermediate,pos;
    return QValidator.Invalid,pos;

class ValueTypeEditor (QWidget):
  ValueTypes = (bool,int,float,complex,str);
  def __init__ (self,*args):
    QWidget.__init__(self,*args);
    lo = QHBoxLayout(self);
    lo.setContentsMargins(0,0,0,0);
    lo.setSpacing(5);
    # type selector
    self.wtypesel = QComboBox(self);
    for i,tp in enumerate(self.ValueTypes):
      self.wtypesel.addItem(tp.__name__);
    QObject.connect(self.wtypesel,SIGNAL("activated(int)"),self._selectTypeNum);
    typesel_lab = QLabel("&Type:",self);
    typesel_lab.setBuddy(self.wtypesel);
    lo.addWidget(typesel_lab,0);
    lo.addWidget(self.wtypesel,0);
    self.wvalue = QLineEdit(self);
    self.wvalue_lab = QLabel("&Value:",self);
    self.wvalue_lab.setBuddy(self.wvalue);
    self.wbool = QComboBox(self);
    self.wbool.addItems(["false","true"]);
    self.wbool.setCurrentIndex(1);
    lo.addWidget(self.wvalue_lab,0);
    lo.addWidget(self.wvalue,1);
    lo.addWidget(self.wbool,1);
    self.wvalue.hide();
    # make input validators
    self._validators = {int:QIntValidator(self),float:QDoubleValidator(self) };
    # select bool type initially
    self._selectTypeNum(0);

  def _selectTypeNum (self,index):
    tp = self.ValueTypes[index];
    self.wbool.setShown(tp is bool);
    self.wvalue.setShown(tp is not bool);
    self.wvalue_lab.setBuddy(self.wbool if tp is bool else self.wvalue);
    self.wvalue.setValidator(self._validators.get(tp,None));

  def setValue (self,value):
    """Sets current value""";
    for i,tp in enumerate(self.ValueTypes):
      if isinstance(value,tp):
        self.wtypesel.setCurrentIndex(i);
        self._selectTypeNum(i);
        if tp is bool:
          self.wbool.setCurrentIndex(1 if value else 0);
        else:
          self.wvalue.setText(str(value));
        return;
    # unknown value: set bool
    self.setValue(True);

  def getValue (self):
    """Returns current value, or None if no legal value is set""";
    tp = self.ValueTypes[self.wtypesel.currentIndex()];
    if tp is bool:
      return bool(self.wbool.currentIndex());
    else:
      try:
        return tp(self.wvalue.text());
      except:
        print "Error converting input to type ",tp.__name__;
        traceback.print_exc();
        return None;


class FileSelector (QWidget):
  """A FileSelector is a one-line widget for selecting a file.""";
  def __init__ (self,parent,label,filename=None,dialog_label=None,file_types=None,default_suffix=None,file_mode=QFileDialog.AnyFile):
    QWidget.__init__(self,parent);
    lo = QHBoxLayout(self);
    lo.setContentsMargins(0,0,0,0);
    lo.setSpacing(5);
    # label
    lab = QLabel(label,self);
    lo.addWidget(lab,0);
    # text field
    self.wfname = QLineEdit(self);
    self.wfname.setReadOnly(True);
    self.setFilename(filename);
    lo.addWidget(self.wfname,1);
    # selector
    wsel = QToolButton(self);
    wsel.setText("Choose...");
    QObject.connect(wsel,SIGNAL("clicked()"),self._chooseFile);
    lo.addWidget(wsel,0);
    # other init
    self._file_dialog = None;
    self._dialog_label = dialog_label or label;
    self._file_types = file_types or "All files (*)";
    self._file_mode = file_mode;
    self._default_suffix = default_suffix;
    self._dir = None;

  def _chooseFile (self):
    if self._file_dialog is None:
      dialog = self._file_dialog = QFileDialog(self,self._dialog_label,".",self._file_types);
      if self._default_suffix:
        dialog.setDefaultSuffix(self._default_suffix);
      dialog.setFileMode(self._file_mode);
      dialog.setModal(True);
      if self._dir is not None:
        dialog.setDirectory(self._dir);
      QObject.connect(dialog,SIGNAL("filesSelected(const QStringList &)"),self.setFilename);
    return self._file_dialog.exec_();

  def setFilename (self,filename):
    if isinstance(filename,QStringList):
      filename = filename[0];
    filename = (filename and str(filename) ) or '';
    self.wfname.setText(filename);
    self.emit(SIGNAL("valid"),bool(filename));
    self.emit(SIGNAL("filenameSelected"),filename);
    
  def setDirectory (self,directory):
    self._dir = directory;
    if self._file_dialog is not None:
      self._file_dialog.setDirectory(directory);

  def filename (self):
    return str(self.wfname.text());

  def isValid (self):
    return bool(self.filename());



class AddTagDialog (QDialog):
  def __init__ (self,parent,modal=True,flags=Qt.WindowFlags()):
    QDialog.__init__(self,parent,flags);
    self.setModal(modal);
    self.setWindowTitle("Add Tag");
    lo = QVBoxLayout(self);
    lo.setMargin(10);
    lo.setSpacing(5);
    # tag selector
    lo1 = QHBoxLayout();
    lo.addLayout(lo1);
    lo1.setSpacing(5);
    self.wtagsel = QComboBox(self);
    self.wtagsel.setEditable(True);
    wtagsel_lbl = QLabel("&Tag:",self);
    wtagsel_lbl.setBuddy(self.wtagsel);
    lo1.addWidget(wtagsel_lbl,0);
    lo1.addWidget(self.wtagsel,1);
    QObject.connect(self.wtagsel,SIGNAL("activated(int)"),self._check_tag);
    QObject.connect(self.wtagsel,SIGNAL("editTextChanged(const QString &)"),self._check_tag_text);
    # value editor
    self.valedit = ValueTypeEditor(self);
    lo.addWidget(self.valedit);
    # buttons
    lo.addSpacing(10);
    lo2 = QHBoxLayout();
    lo.addLayout(lo2);
    lo2.setContentsMargins(0,0,0,0);
    lo2.setMargin(5);
    self.wokbtn = QPushButton("OK",self);
    self.wokbtn.setMinimumWidth(128);
    QObject.connect(self.wokbtn,SIGNAL("clicked()"),self.accept);
    self.wokbtn.setEnabled(False);
    cancelbtn = QPushButton("Cancel",self);
    cancelbtn.setMinimumWidth(128);
    QObject.connect(cancelbtn,SIGNAL("clicked()"),self.reject);
    lo2.addWidget(self.wokbtn);
    lo2.addStretch(1);
    lo2.addWidget(cancelbtn);
    self.setMinimumWidth(384);

  def setTags (self,tagnames):
    self.wtagsel.clear();
    self.wtagsel.addItems(list(tagnames));
    self.wtagsel.addItem("");
    self.wtagsel.setCurrentIndex(len(tagnames));

  def setValue (self,value):
    self.valedit.setValue(value);

  def _check_tag (self,tag):
    self.wokbtn.setEnabled(True);

  def _check_tag_text (self,text):
    self.wokbtn.setEnabled(bool(str(text)!=""));
    
  def accept (self):
    """When dialog is accepted with a default (bool) tag type,
    check if the user hasn't entered a name=value entry in the tag name field.
    This is a common mistake, and should be treated as a shortcut for setting string tags.""";
    if isinstance(self.valedit.getValue(),bool):
      tagval = str(self.wtagsel.currentText()).split("=",1);
      if len(tagval) > 1:
#        print tagval;
        if QMessageBox.warning(self,
            "Set a string tag instead?","""<P>You have included an "=" sign in the tag name. 
            Perhaps you actually mean to set tag "%s" to the string value "%s"?</P>"""%tuple(tagval),
                  QMessageBox.Yes|QMessageBox.No,QMessageBox.Yes) == QMessageBox.No:
          return;
        self.wtagsel.setEditText(tagval[0]);
        self.valedit.setValue(tagval[1]);
    return QDialog.accept(self);

  def getTag (self):
    return str(self.wtagsel.currentText()),self.valedit.getValue();

class SelectTagsDialog (QDialog):
  def __init__ (self,parent,modal=True,flags=Qt.WindowFlags(),caption="Select Tags",ok_button="Select"):
    QDialog.__init__(self,parent,flags);
    self.setModal(modal);
    self.setWindowTitle(caption);
    lo = QVBoxLayout(self);
    lo.setMargin(10);
    lo.setSpacing(5);
    # tag selector
    self.wtagsel = QListWidget(self);
    lo.addWidget(self.wtagsel);
#    self.wtagsel.setColumnMode(QListBox.FitToWidth);
    self.wtagsel.setSelectionMode(QListWidget.MultiSelection);
    QObject.connect(self.wtagsel,SIGNAL("itemSelectionChanged()"),self._check_tag);
    # buttons
    lo.addSpacing(10);
    lo2 = QHBoxLayout();
    lo.addLayout(lo2);
    lo2.setContentsMargins(0,0,0,0);
    lo2.setMargin(5);
    self.wokbtn = QPushButton(ok_button,self);
    self.wokbtn.setMinimumWidth(128);
    QObject.connect(self.wokbtn,SIGNAL("clicked()"),self.accept);
    self.wokbtn.setEnabled(False);
    cancelbtn = QPushButton("Cancel",self);
    cancelbtn.setMinimumWidth(128);
    QObject.connect(cancelbtn,SIGNAL("clicked()"),self.reject);
    lo2.addWidget(self.wokbtn);
    lo2.addStretch(1);
    lo2.addWidget(cancelbtn);
    self.setMinimumWidth(384);
    self._tagnames = [];

  def setTags (self,tagnames):
    self._tagnames = tagnames;
    self.wtagsel.clear();
    self.wtagsel.insertItems(0,list(tagnames));

  def _check_tag (self):
    for i in range(len(self._tagnames)):
      if self.wtagsel.item(i).isSelected():
        self.wokbtn.setEnabled(True);
        return;
    else:
      self.wokbtn.setEnabled(False);

  def getSelectedTags (self):
    return [ tag for i,tag in enumerate(self._tagnames) if self.wtagsel.item(i).isSelected() ];



# -*- coding: utf-8 -*-
#
#% $Id$
#
#
# Copyright (C) 2002-2011
# The MeqTree Foundation &
# ASTRON (Netherlands Foundation for Research in Astronomy)
# P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>,
# or write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

import math
from PyQt4.Qt import *

import Kittens.widgets
import Kittens.utils
from Kittens.utils import PersistentCurrier
from Kittens.widgets import BusyIndicator

from Tigger.Models import ModelClasses,PlotStyles
from Tigger.Models.SkyModel import SkyModel

_verbosity = Kittens.utils.verbosity(name="tw");
dprint = _verbosity.dprint;
dprintf = _verbosity.dprintf;

ViewColumns = [ "name","RA","RA err","Dec","Dec err","r","type",
  "Iapp","I","I err","Q","Q err","U","U err","V","V err","RM","RM err","spi","spi err",
  "shape","shape err","tags" ];

for icol,col in enumerate(ViewColumns):
    globals()["Column%s"%col.capitalize().replace(" ","_")] = icol;
NumColumns = len(ViewColumns);

DEG = math.pi/180;

# Qt-4.6 and up (PyQt 4.7 and up) has very slow QTreeWidgetItem updates, determine version here
from PyQt4 import QtCore
_SLOW_QTREEWIDGETITEM = QtCore.PYQT_VERSION_STR >= '4.7';

class SkyModelTreeWidget (Kittens.widgets.ClickableTreeWidget):
  """This implements a QTreeWidget for sky models""";
  def __init__ (self,*args):
    Kittens.widgets.ClickableTreeWidget.__init__(self,*args);
    self._currier = PersistentCurrier();
    self.model = None;
    # insert columns
    self.setHeaderLabels(ViewColumns);
    self.headerItem().setText(ColumnIapp,"I(app)");
    self.header().setMovable(False);
    self.header().setClickable(True);
    self.setSortingEnabled(True);
    self.setRootIsDecorated(False);
    self.setEditTriggers(QAbstractItemView.AllEditTriggers);
    self.setMouseTracking(True);
    # set column width modes
    for icol in range(NumColumns-1):
      self.header().setResizeMode(icol,QHeaderView.ResizeToContents);
    self.header().setStretchLastSection(True);
    ## self.setTextAlignment(ColumnR,Qt.AlignRight);
    ## self.setTextAlignment(ColumnType,Qt.AlignHCenter);
    # _column_enabled[i] is True if column is available in the model.
    # _column_show[i] is True if column is currently being shown (via a view control)
    self._column_enabled = [True]*NumColumns;
    self._column_shown    = [True]*NumColumns;
    # other listview init
    self.header().show();
    self.setSelectionMode(QTreeWidget.ExtendedSelection);
    self.setAllColumnsShowFocus(True);
    ## self.setShowToolTips(True);
    self._updating_selection = False;
    self.setRootIsDecorated(False);
    # connect signals to track selected sources
    QObject.connect(self,SIGNAL("itemSelectionChanged()"),self._selectionChanged);
    QObject.connect(self,SIGNAL("itemEntered(QTreeWidgetItem*,int)"),self._itemHighlighted);
    # add "View" controls for different column categories
    self._column_views = [];
    self._column_widths = {};
    self.addColumnCategory("Position",[ColumnRa,ColumnDec]);
    self.addColumnCategory("Position errors",[ColumnRa_err,ColumnDec_err],False);
    self.addColumnCategory("Type",[ColumnType]);
    self.addColumnCategory("Flux",[ColumnIapp,ColumnI]);
    self.addColumnCategory("Flux errors",[ColumnI_err],False);
    self.addColumnCategory("Polarization",[ColumnQ,ColumnU,ColumnV,ColumnRm]);
    self.addColumnCategory("Polarization errors",[ColumnQ_err,ColumnU_err,ColumnV_err,ColumnRm_err],False);
    self.addColumnCategory("Spectrum",[ColumnSpi]);
    self.addColumnCategory("Spectrum errors",[ColumnSpi_err],False);
    self.addColumnCategory("Shape",[ColumnShape]);
    self.addColumnCategory("Shape errors",[ColumnShape_err],False);
    self.addColumnCategory("Tags",[ColumnTags]);

  def _showColumn (self,col,show=True):
    """Shows or hides the specified column.
    (When hiding, saves width of column to internal array so that it can be restored properly.)"""
    hdr = self.header();
    hdr.setSectionHidden(col,not show)
    if show:
      if not hdr.sectionSize(col):
        hdr.resizeSection(col,self._column_widths[col]);
        hdr.setResizeMode(col,QHeaderView.ResizeToContents);
    else:
      if hdr.sectionSize(col):
        self._column_widths[col] = hdr.sectionSize(col);

  def _enableColumn (self,column,enable=True):
    busy = BusyIndicator();
    self._column_enabled[column] = enable;
    self._showColumn(column,enable and self._column_shown[column]);

  def _showColumnCategory (self,columns,show):
    busy = BusyIndicator();
    for col in columns:
      self._column_shown[col] = show;
      self._showColumn(col,self._column_enabled[col] and show);

  def _selectionChanged (self):
    if self._updating_selection:
      return;
    for item in self.iterator():
      item._src.select(item.isSelected());
    self.model.emitSelection(origin=self);

  def _itemHighlighted (self,item,col):
    dprint(3,"highlighting",item._src.name);
    self.model.setCurrentSource(item._src,origin=self);

  def viewportEvent (self,event):
    if event.type() in (QEvent.Leave,QEvent.FocusOut) and self.model:
      self.model.setCurrentSource(None,origin=self);
    return QTreeWidget.viewportEvent(self,event);

  def addColumnCategory (self,name,columns,visible=True):
    qa = QAction(name,self);
    qa.setCheckable(True);
    qa.setChecked(visible);
    if not visible:
      self._showColumnCategory(columns,False)
    QObject.connect(qa,SIGNAL("toggled(bool)"),self._currier.curry(self._showColumnCategory,columns));
    self._column_views.append((name,qa,columns));

  def clear (self):
    Kittens.widgets.ClickableTreeWidget.clear(self);
    self.model = None;
    self._itemdict = {};

  def setModel (self,model):
    self.model = model;
    self._refreshModel(SkyModel.UpdateAll);
    self.model.connect("changeCurrentSource",self._updateCurrentSource);
    self.model.connect("changeGroupingVisibility",self.changeGroupingVisibility);
    self.model.connect("selected",self._updateModelSelection);
    self.model.connect("updated",self._refreshModel);

  def _refreshModel (self,what=SkyModel.UpdateAll,origin=None):
    if origin is self or not what&(SkyModel.UpdateSourceList|SkyModel.UpdateSourceContent):
      return;
    # if only selection was changed, take shortcut
    if what&SkyModel.UpdateSelectionOnly:
      dprint(2,"model update -- selection only");
      return self._refreshSelectedItems(origin);
    busy = BusyIndicator();
    # else repopulate widget completely
    dprint(2,"model update -- complete");
    Kittens.widgets.ClickableTreeWidget.clear(self);
    dprint(2,"creating model items");
    items = [ SkyModelTreeWidgetItem(src) for src in self.model.sources ];
    self._itemdict = dict(zip([src.name for src in self.model.sources],items));
    dprint(2,"adding to tree widget");
    self.addTopLevelItems(items);
    self.header().updateGeometry();
    # show/hide columns based on tag availability
    self._enableColumn(ColumnIapp,'Iapp' in self.model.tagnames);
    self._enableColumn(ColumnR,'r' in self.model.tagnames);
    dprint(2,"re-sorting");
    self.sortItems(('Iapp' in self.model.tagnames and ColumnIapp) or ColumnI,Qt.DescendingOrder);
    busy = None;

  def addColumnViewActionsTo (self,menu):
    for name,qa,columns in self._column_views:
      menu.addAction(qa);

  def _updateCurrentSource (self,src,src0=None,origin=None):
    # if origin is self:
    # return;
    # dehighlight old item
    item = src0 and self._itemdict.get(src0.name);
    if item:
      item.setHighlighted(False);
    # scroll to new item, if found
    item = src and self._itemdict.get(src.name);
    if item:
      item.setHighlighted(True,origin is not self);
      if origin is not self:
        self.scrollToItem(item);

  def _updateModelSelection (self,nsel,origin=None):
    """This is called when some other widget (origin!=self) changes the set of selected model sources""";
    if origin is self:
      return;
    self._updating_selection = True;
## this is very slow because of setSelected()
#    for item in self.iterator():
#     item.setSelected(item._src.selected);
    selection = QItemSelection();
    for item in self.iterator():
      if item._src.selected:
        selection.append(QItemSelectionRange(self.indexFromItem(item,0),self.indexFromItem(item,self.columnCount()-1)));
    self.selectionModel().select(selection,QItemSelectionModel.ClearAndSelect);
    self.changeGroupingVisibility(None,origin=origin);
    self._updating_selection = False;

  def _refreshSelectedItems (self,origin=None):
    busy = BusyIndicator();
    dprint(3,"refreshing selected items");
    for item in self.iterator():
      if item.isSelected():
        dprint(4,"resetting item",item._src.name);
        item.setSource(item._src);
    dprint(3,"refreshing selected items done");
    busy = None;

  def changeGroupingVisibility (self,group,origin=None):
    if origin is self:
      return;
    for item in self.iterator():
      # collect show_list values from groupings to which this source belongs (default group excepted)
      show = [ group.style.show_list for group in self.model.groupings if group is not self.model.defgroup and group.func(item._src) ];
      # if at least one group is showing explicitly, show
      # else if at least one group is hiding explicitly, hide
      # else use default setting
      if max(show) == PlotStyles.ShowAlways:
        visible = True;
      elif min(show) == PlotStyles.ShowNot:
        visible = False;
      else:
        visible = bool(self.model.defgroup.style.show_list);
      # set visibility accordingly
      item.setHidden(not visible);

  TagsWithOwnColumn = set(["Iapp","r"]);

class SkyModelTreeWidgetItem (QTreeWidgetItem):

  _fonts = None;
  @staticmethod
  def _initFonts ():
    """Initializes fonts on fitrst call""";
    if SkyModelTreeWidgetItem._fonts is None:
      stdfont = QApplication.font();
      boldfont = QFont(stdfont);
      boldfont.setBold(True);
      SkyModelTreeWidgetItem._fonts = [ stdfont,boldfont ];
      SkyModelTreeWidgetItem._fontmetrics = QFontMetrics(boldfont);

  def __init__ (self,src,*args):
    QTreeWidgetItem.__init__(self,*args);
    self._src = src;
    # fonts
    self._initFonts();
    # array of actual (i.e. numeric) column values
    self._values = [None]*NumColumns;
    # set text alignment
    for icol in range(NumColumns):
      self.setTextAlignment(icol,Qt.AlignLeft);
    self.setTextAlignment(ColumnR,Qt.AlignRight);
    self.setTextAlignment(ColumnType,Qt.AlignHCenter);
    # setup source
    self._highlighted = self._highlighted_visual = False;
    self.setSource(src);

  def setHighlighted (self,highlighted=True,visual=False):
#    global _SLOW_QTREEWIDGETITEM;
#    if 1: # not _SLOW_QTREEWIDGETITEM:
    visual = True;
    dprint(3,self._src.name,"highlighted",highlighted,visual);
    if highlighted != self._highlighted:
#      brush = QApplication.palette().alternateBase() if highlighted else QApplication.palette().base();
#      for col in range(self.columnCount()):
#        self.setBackground(col,brush);
      if highlighted and visual:
        self.setFont(0,self._fonts[1]);
      elif not highlighted and self._highlighted_visual:
        self.setFont(0,self._fonts[0]);
      self._highlighted = highlighted;
      self._highlighted_visual = visual;

  @staticmethod
  def _angErrToStr (value):
    """helper method: converts angular error to string representation in deg or arcmin or arcsec""";
    arcsec = (value/DEG)*3600;
    if arcsec < 60:
      return unichr(0xB1)+"%.2g\""%arcsec;
    elif arcsec < 3600:
      return unichr(0xB1)+"%.2f'"%(arcsec*60);
    else:
      return unichr(0xB1)+"%.2f%s"%(arcsec*3600,unichr(0xB0));


  def setSource (self,src):
    # name
    dprint(3,"setSource 1",src.name);
    self.setColumn(ColumnName,src.name);
    self.setSizeHint(0,QSize(self._fontmetrics.width("x"+src.name),0));
    # coordinates
    self.setColumn(ColumnRa,src.pos.ra,"%2dh%02dm%05.2fs"%src.pos.ra_hms());
    self.setColumn(ColumnDec,src.pos.dec,("%s%2d"+unichr(0xB0)+"%02d'%05.2f\"")%
        src.pos.dec_sdms());
    if src.pos.ra_err is not None:
      self.setColumn(ColumnRa_err,src.pos.ra_err,self._angErrToStr(src.pos.ra_err));
    if src.pos.dec_err is not None:
      self.setColumn(ColumnDec_err,src.pos.dec_err,self._angErrToStr(src.pos.dec_err));
    if hasattr(src,'r'):
      self.setColumn(ColumnR,src.r,"%.1f'"%(src.r*180*60/math.pi));
    # type
    self.setColumn(ColumnType,src.typecode);
    # flux
    if hasattr(src,'Iapp'):
      self.setColumn(ColumnIapp,src.Iapp,"%.3g"%src.Iapp);
    for stokes in "IQUV":
      stk = getattr(src.flux,stokes,None);
      stk_err = getattr(src.flux,stokes+"_err",None);
      if stk is not None:
        self.setColumn(globals()['Column'+stokes],stk,"%.3g"%stk);
      if stk_err is not None:
        self.setColumn(globals()['Column'+stokes+"_err"],stk_err,unichr(0xB1)+"%.2g"%stk_err);
    if hasattr(src.flux,'rm'):
      self.setColumn(ColumnRm,src.flux.rm,"%.2f"%src.flux.rm);
      if hasattr(src.flux,'rm_err'):
        self.setColumn(ColumnRm_err,src.flux.rm_err,unichr(0xB1)+"%.2f"%src.flux.rm);
    # spi
    if isinstance(src.spectrum,ModelClasses.SpectralIndex):
      spi = getattr(src.spectrum,'spi',0);
      if not isinstance(spi,(list,tuple)):
        spi = [spi];
      spi = ",".join([ "%.2f"%x for x in spi]);
      self.setColumn(ColumnSpi,src.spectrum.spi,spi );
      spierr = getattr(src.spectrum,'spi_err',None);
      if spierr is not None:
        if not isinstance(spierr,(list,tuple)):
          spierr = [spierr];
        spierr = ",".join([ "%.2f"%x for x in spierr]);
        self.setColumn(ColumnSpi_err,src.spectrum.spi_err,unichr(0xB1)+spierr);
    # shape
    shape = getattr(src,'shape',None);
    if isinstance(shape,ModelClasses.ModelItem):
      shapeval = shape.getShape();
      shapestr = shape.strDesc(delimiters=('"',unichr(0xD7),unichr(0x21BA),unichr(0xB0)));
      self.setColumn(ColumnShape,shapeval,shapestr);
      errval = shape.getShapeErr();
      if errval:
        errstr = shape.strDescErr(delimiters=('"',unichr(0xD7),unichr(0x21BA),unichr(0xB0)));
        self.setColumn(ColumnShape_err,errval,unichr(0xB1)+errstr);
    dprint(3,"setSource 3",src.name);
    # Tags. Tags are all extra attributes that do not have a dedicated column (i.e. not Iapp or r), and do not start
    # with "_" (which is reserved for internal attributes)

    ## the complexity below seems entirely unnecessary, since sorting the tag strings automatically puts "_" first,
    ## "-" second, and alphabet afterwards

    #truetags = [];
    #falsetags = [];
    #othertags = [];
    #for attr,val in src.getExtraAttributes():
      #if attr[0] != "_" and attr not in SkyModelTreeWidget.TagsWithOwnColumn:
        #if val is False:
          #falsetags.append("-"+attr);
        #elif val is True:
          #truetags.append("+"+attr);
        #else:
          #othertags.append("%s=%s"%(attr,str(val)));
    #for tags in truetags,falsetags,othertags:
      #tags.sort();
    #self.setColumn(ColumnTags,tags," ".join(truetags+falsetags+othertags));

    # so instead:
    tags = [ "+"+attr if val is True else "-"+attr if val is False else "%s=%s"%(attr,str(val))
      for attr,val in src.getExtraAttributes()
      if attr[0] != "_" and attr not in SkyModelTreeWidget.TagsWithOwnColumn ];
    tagstr = " ".join(sorted(tags));
    dprint(3,"setSource 4",src.name);
    self.setColumn(ColumnTags,tags,tagstr);
    dprint(3,"setSource 5",src.name);
    dprint(3,"setSource done",src.name);

  def setColumn (self,icol,value,text=None):
    """helper function to set the value of a column""";
    if text is None:
      text = str(value);
    self.setText(icol,text);
    self._values[icol] = value;

  def __lt__ (self,other):
    icol = self.treeWidget().sortColumn();
    if isinstance(other,SkyModelTreeWidgetItem):
      return self._values[icol] < other._values[icol];
    else:
      return self.text(icol) < other.text(icol);

  def __ge__ (self,other):
    return other < self;


class ModelGroupsTable (QWidget):
  EditableAttrs = [ attr for attr in PlotStyles.StyleAttributes if attr in PlotStyles.StyleAttributeOptions ];
  ColList = 3;
  ColPlot = 4;
  ColApply = 5;
  AttrByCol = dict([(i+6,attr) for i,attr in enumerate(EditableAttrs)]);

  def __init__ (self,parent,*args):
    QWidget.__init__(self,parent,*args);
    self.setSizePolicy(QSizePolicy.Expanding,QSizePolicy.Expanding);
    lo = QVBoxLayout(self);
    lo.setContentsMargins(0,0,0,0);
    lo1 = QHBoxLayout();
    lo.addLayout(lo1);
    lo1.setContentsMargins(0,0,0,0);
    lbl = QLabel(QString("<nobr><b>Source groupings:</b></nobr>"),self);
    lo1.addWidget(lbl,0);
    lo1.addStretch(1);
    # add show/hide button
    self._showattrbtn = QPushButton(self);
    self._showattrbtn.setMinimumWidth(256);
    lo1.addWidget(self._showattrbtn,0);
    lo1.addStretch();
    QObject.connect(self._showattrbtn,SIGNAL("clicked()"),self._togglePlotControlsVisibility);
    # add table
    self.table  = QTableWidget(self);
    lo.addWidget(self.table);
    QObject.connect(self.table,SIGNAL("cellChanged(int,int)"),self._valueChanged);
    self.table.setSelectionMode(QTableWidget.NoSelection);
    # setup basic columns
    self.table.setColumnCount(6+len(self.EditableAttrs));
    for i,label in enumerate(("grouping","total","selection","list","plot","style")):
      self.table.setHorizontalHeaderItem(i,QTableWidgetItem(label));
    self.table.horizontalHeader().setSectionHidden(self.ColApply,True);
    # setup columns for editable grouping attributes
    for i,attr in self.AttrByCol.iteritems():
      self.table.setHorizontalHeaderItem(i,QTableWidgetItem(PlotStyles.StyleAttributeLabels[attr]));
      self.table.horizontalHeader().setSectionHidden(i,True);
    self.table.verticalHeader().hide();
    # other internal init
    self._attrs_shown = False;
    self._togglePlotControlsVisibility();
    self.model = None;
    self._setting_model = False;
    self._currier = PersistentCurrier();
    # row of 'selected' grouping
    self._irow_selgroup = 0;

  def clear (self):
    self.table.setRowCount(0);
    self.model = None;

  # setup mappings from the group.show_plot attribute to check state
  ShowAttrToCheckState = {  PlotStyles.ShowNot:Qt.Unchecked,
                            PlotStyles.ShowDefault:Qt.PartiallyChecked,
                            PlotStyles.ShowAlways:Qt.Checked };
  CheckStateToShowAttr = dict([(val,key) for key,val in ShowAttrToCheckState.iteritems ()]);

  def _makeCheckItem (self,name,group,attr):
    item = QTableWidgetItem(name);
    if group is self.model.defgroup:
      item.setFlags(Qt.ItemIsEnabled|Qt.ItemIsUserCheckable);
      item.setCheckState(Qt.Checked if getattr(group.style,attr) else Qt.Unchecked);
    else:
      item.setFlags(Qt.ItemIsEnabled|Qt.ItemIsUserCheckable|Qt.ItemIsTristate);
      item.setCheckState(self.ShowAttrToCheckState[getattr(group.style,attr)]);
    return item;

  def _updateModel (self,what=SkyModel.UpdateAll,origin=None):
    if origin is self or not what&(SkyModel.UpdateTags|SkyModel.UpdateGroupStyle):
      return;
    model = self.model;
    self._setting_model= True;  # to ignore cellChanged() signals (in valueChanged())
    # _item_cb is a dict (with row,col keys) containing the widgets (CheckBoxes ComboBoxes) per each cell
    self._item_cb = {};
    # lists of "list" and "plot" checkboxes per each grouping (excepting the default grouping); each entry is an (row,col,item) tuple.
    # used as argument to self._showControls()
    self._list_controls = [];
    self._plot_controls = [];
    # list of selection callbacks (to which signals are connected)
    self._callbacks = [];
    # set requisite number of rows,and start filling
    self.table.setRowCount(len(model.groupings));
    for irow,group in enumerate(model.groupings):
      self.table.setItem(irow,0,QTableWidgetItem(group.name));
      if group is model.selgroup:
        self._irow_selgroup = irow;
      # total # source in group: skip for "current"
      if group is not model.curgroup:
        self.table.setItem(irow,1,QTableWidgetItem(str(group.total)));
      # selection controls: skip for current and selection
      if group not in (model.curgroup,model.selgroup):
        btns = QWidget();
        lo = QHBoxLayout(btns);
        lo.setContentsMargins(0,0,0,0);
        lo.setSpacing(0);
        # make selector buttons (depending on which group we're in)
        if group is model.defgroup:
          Buttons = (
            ("+",lambda src,grp=group:True,"select all sources"),
            ("-",lambda src,grp=group:False,"unselect all sources") );
        else:
          Buttons = (
            ("=",lambda src,grp=group:grp.func(src),"select only this grouping"),
            ("+",lambda src,grp=group:src.selected or grp.func(src),"add grouping to selection"),
            ("-",lambda src,grp=group:src.selected and not grp.func(src),"remove grouping from selection"),
            ("&&",lambda src,grp=group:src.selected and grp.func(src),"intersect selection with grouping"));
        lo.addStretch(1);
        for label,predicate,tooltip in Buttons:
          btn = QToolButton(btns);
          btn.setText(label);
          btn.setMinimumWidth(24);
          btn.setMaximumWidth(24);
          btn.setToolTip(tooltip);
          lo.addWidget(btn);
          # add callback
          QObject.connect(btn,SIGNAL("clicked()"),self._currier.curry(self.selectSources,predicate));
        lo.addStretch(1);
        self.table.setCellWidget(irow,2,btns);
      # "list" checkbox (not for current and selected groupings: these are always listed)
      if group not in (model.curgroup,model.selgroup):
        item = self._makeCheckItem("",group,"show_list");
        self.table.setItem(irow,self.ColList,item);
        item.setToolTip("""<P>If checked, sources in this grouping will be listed in the source table. If un-checked, sources will be
            excluded from the table. If partially checked, then the default list/no list setting of "all sources" will be in effect.
            </P>""");
      # "plot" checkbox (not for the current grouping, since that's always plotted)
      if group is not model.curgroup:
        item = self._makeCheckItem("",group,"show_plot");
        self.table.setItem(irow,self.ColPlot,item);
        item.setToolTip("""<P>If checked, sources in this grouping will be included in the plot. If un-checked, sources will be
            excluded from the plot. If partially checked, then the default plot/no plot setting of "all sources" will be in effect.
            </P>""");
      # custom style control
      # for default, current and selected, this is just a text label
      if group is model.defgroup:
        item = QTableWidgetItem("default:");
        item.setTextAlignment(Qt.AlignRight|Qt.AlignVCenter);
        item.setToolTip("""<P>This is the default plot style used for all sources for which a custom grouping style is not selected.</P>""");
        self.table.setItem(irow,self.ColApply,item);
      elif group is model.curgroup:
        item = QTableWidgetItem("");
        item.setTextAlignment(Qt.AlignRight|Qt.AlignVCenter);
        item.setToolTip("""<P>This is the plot style used for the highlighted source, if any.</P>""");
        self.table.setItem(irow,self.ColApply,item);
      elif group is model.selgroup:
        item = QTableWidgetItem("");
        item.setTextAlignment(Qt.AlignRight|Qt.AlignVCenter);
        item.setToolTip("""<P>This is the plot style used for the currently selected sources.</P>""");
        self.table.setItem(irow,self.ColApply,item);
      # for the rest, a combobox with custom priorities
      else:
        cb = QComboBox();
        cb.addItems(["default"]+["custom %d"%p for p in range(1,10)]);
        index = max(0,min(group.style.apply,9));
#        dprint(0,group.name,"apply",index);
        cb.setCurrentIndex(index);
        QObject.connect(cb,SIGNAL("activated(int)"),self._currier.xcurry(self._valueChanged,(irow,self.ColApply)));
        self.table.setCellWidget(irow,self.ColApply,cb);
        cb.setToolTip("""<P>This controls whether sources within this group are plotted with a customized
            plot style. Customized styles have numeric priority; if a source belongs to multiple groups, then
            the style with the lowest priority takes precedence.<P>""");
      # attribute comboboxes
      for icol,attr in self.AttrByCol.iteritems():
        # get list of options for this style attribute. If dealing with first grouping (i==0), which is
        # the "all sources" grouping, then remove the "default" option (which is always first in the list)
        options = PlotStyles.StyleAttributeOptions[attr];
        if irow == 0:
          options = options[1:];
        # make combobox
        cb = QComboBox();
        cb.addItems(map(str,options));
        # the "label" option is also editable
        if attr == "label":
          cb.setEditable(True);
        try:
          index = options.index(getattr(group.style,attr));
          cb.setCurrentIndex(index);
        except ValueError:
          cb.setEditText(str(getattr(group.style,attr)));
        slot = self._currier.xcurry(self._valueChanged,(irow,icol));
        QObject.connect(cb,SIGNAL("activated(int)"),slot);
        QObject.connect(cb,SIGNAL("editTextChanged(const QString &)"),slot);
        cb.setEnabled(group is model.defgroup or group.style.apply);
        self.table.setCellWidget(irow,icol,cb);
        label = attr;
        if irow:
          cb.setToolTip("""<P>This is the %s used to plot sources in this group, when a "custom" style for the group
          is enabled via the style control.<P>"""%label);
        else:
          cb.setToolTip("<P>This is the default %s used for all sources for which a custom style is not specified below.<P>"%label);
    self.table.resizeColumnsToContents();
    # re-enable processing of cellChanged() signals
    self._setting_model= False;

  def setModel (self,model):
    self.model = model;
    self.model.connect("updated",self._updateModel);
    self.model.connect("selected",self.updateModelSelection);
    self._updateModel(SkyModel.UpdateAll);

  def _valueChanged (self,row,col):
    """Called when a cell has been edited""";
    if self._setting_model:
      return;
    group = self.model.groupings[row];
    item = self.table.item(row,col);
    if col == self.ColList:
      if group is not self.model.defgroup:
        # tri-state items go from unchecked to checked when user clicks them. Make them partially checked instead.
        if group.style.show_list == PlotStyles.ShowNot and item.checkState() == Qt.Checked:
          item.setCheckState(Qt.PartiallyChecked);
      group.style.show_list = self.CheckStateToShowAttr[item.checkState()];
      self.model.emitChangeGroupingVisibility(group,origin=self);
      return;
    elif col == self.ColPlot:
      if group is not self.model.defgroup:
        # tri-state items go from unchecked to checked by default. Make them partially checked instead.
        if group.style.show_plot == PlotStyles.ShowNot and item.checkState() == Qt.Checked:
          item.setCheckState(Qt.PartiallyChecked);
      group.style.show_plot = self.CheckStateToShowAttr[item.checkState()];
    elif col == self.ColApply:
      group.style.apply = self.table.cellWidget(row,col).currentIndex();
      # enable/disable editable cells
      for j in self.AttrByCol.keys():
        item1 = self.table.item(row,j);
        if item1:
          fl = item1.flags()&~Qt.ItemIsEnabled;
          if group.style.apply:
            fl |= Qt.ItemIsEnabled;
          item1.setFlags(fl);
        cw = self.table.cellWidget(row,j);
        cw and cw.setEnabled(group.style.apply);
    elif col in self.AttrByCol:
      cb = self.table.cellWidget(row,col);
      txt = str(cb.currentText());
      attr = self.AttrByCol[col];
      if txt == "default":
        setattr(group.style,attr,PlotStyles.DefaultValue);
      else:
        setattr(group.style,attr,PlotStyles.StyleAttributeTypes.get(attr,str)(txt));
    # all other columns: return so we don't emit a signal
    else:
      return;
    # in all cases emit a signal
    self.model.emitChangeGroupingStyle(group,origin=self);

  def selectSources (self,predicate):
    """Selects sources according to predicate(src)"""
    busy = BusyIndicator();
    for src in self.model.sources:
      src.selected = predicate(src);
    self.model.emitSelection(origin=self);
    busy = None;

  def updateModelSelection (self,nsel,origin=None):
    """This is called when some other widget changes the set of selected model sources""";
    self.table.clearSelection();
    if self.model:
      self.table.item(self._irow_selgroup,1).setText(str(nsel));

  def _togglePlotControlsVisibility (self):
    if self._attrs_shown:
      self._attrs_shown = False;
      self.table.hideColumn(self.ColApply);
      for col in self.AttrByCol.iterkeys():
        self.table.hideColumn(col);
      self._showattrbtn.setText("Show plot styles >>");
    else:
      self._attrs_shown = True;
      self.table.showColumn(self.ColApply);
      for col in self.AttrByCol.iterkeys():
        self.table.showColumn(col);
      self._showattrbtn.setText("<< Hide plot styles");





#!/usr/bin/env python

from __future__ import print_function
import os
import sys
from setuptools import setup, find_packages
import warnings

msg = """
Could not detect PyQt4. Install PyQt4 system wide. If you are in
a virtualenv install the vext.pyqt4 package: 

  $ pip install astro-tigger[venv]

"""

try:
    import PyQt4
except ImportError:
    warnings.warn(msg)


__version__ = "1.4.0"

requirements = ['astro_kittens', 'numpy', 'scipy', 'astlib', 'pyfits', 'astro_tigger_lsm' ]

scripts = [
    'TigGUI/tigger',
]

package_data = {'TigGUI': [
    'icons/*.png',
    'tigger.conf',
] }

extras_require = {
    'venv': ['vext.pyqt4'],
}


setup(
    name ="astro-tigger",
    version=__version__,
    packages=find_packages(),
    extras_require=extras_require,
    scripts=scripts,
    package_data=package_data,
    description="yet another FITS image viewer",
    author="Oleg Smirnov",
    author_email="osmirnov@gmail.com",
    url="https://github.com/ska-sa/tigger",
    install_requires=requirements,
)




#
#% $Id$ 
#
#
# Copyright (C) 2002-2011
# The MeqTree Foundation & 
# ASTRON (Netherlands Foundation for Research in Astronomy)
# P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>,
# or write to the Free Software Foundation, Inc., 
# 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

from PyQt4.Qt import *
import math
import os.path
from Kittens.widgets import BusyIndicator
from TigGUI.Widgets import FileSelector

DEG = math.pi/180;


class ExportKarmaDialog (QDialog):
  def __init__ (self,parent,modal=True,flags=Qt.WindowFlags()):
    QDialog.__init__(self,parent,flags);
    self.setModal(modal);
    self.setWindowTitle("Export Karma annotations");
    lo = QVBoxLayout(self);
    lo.setMargin(10);
    lo.setSpacing(5);
    # file selector
    self.wfile = FileSelector(self,label="Filename:",dialog_label="Karma annotations filename",default_suffix="ann",file_types="Karma annotations (*.ann)");
    lo.addWidget(self.wfile);
    # selected sources checkbox
    self.wsel = QCheckBox("selected sources only",self);
    lo.addWidget(self.wsel);
    # OK/cancel buttons
    lo.addSpacing(10);
    lo2 = QHBoxLayout();
    lo.addLayout(lo2);
    lo2.setContentsMargins(0,0,0,0);
    lo2.setMargin(5);
    self.wokbtn = QPushButton("OK",self);
    self.wokbtn.setMinimumWidth(128);
    QObject.connect(self.wokbtn,SIGNAL("clicked()"),self.accept);
    self.wokbtn.setEnabled(False);
    cancelbtn = QPushButton("Cancel",self);
    cancelbtn.setMinimumWidth(128);
    QObject.connect(cancelbtn,SIGNAL("clicked()"),self.reject);
    lo2.addWidget(self.wokbtn);
    lo2.addStretch(1);
    lo2.addWidget(cancelbtn);
    self.setMinimumWidth(384);
    # signals
    QObject.connect(self.wfile,SIGNAL("valid"),self.wokbtn.setEnabled);
    # internal state
    self.qerrmsg = QErrorMessage(self);
    self._model_filename = None;

  def setModel (self,model):
    self.model = model;
    # set the default annotations filename, whenever a new model filename is set
    filename = self.model.filename();
    if filename and filename != self._model_filename:
      self._model_filename = filename;
      self.wfile.setFilename(os.path.splitext(filename)[0] + ".ann");

  def accept (self):
    """Tries to export annotations, and closes the dialog if successful.""";
    try:
      filename = self.wfile.filename();
      if os.path.exists(filename) and QMessageBox.question(self,"Exporting Karma annotations","<P>Overwrite the file %s?</P>"%filename,
                QMessageBox.Yes|QMessageBox.No,QMessageBox.Yes) != QMessageBox.Yes:
          return;
      f = file(self.wfile.filename(),"wt");
      f.write('COORD W\nPA STANDARD\nCOLOR GREEN\nFONT hershey12\n');
      # source list
      if self.wsel.isChecked():
        sources = [ src for src in self.model.sources if src.selected ];
      else:
        sources = self.model.sources;
      # calculate basis size for crosses (TODO: replace min_size with something more sensible, as this value is in degrees)
      brightnesses = [ abs(src.brightness()) for src in sources if src.brightness() != 0 ];
      min_bright = brightnesses and min(brightnesses) ;
      min_size = 0.01;
      # loop over sources
      busy = BusyIndicator();
      for src in sources:
        ra = src.pos.ra/DEG;
        dec = src.pos.dec/DEG;
        # figure out source size
        if src.brightness() and min_bright:
          ysize = (math.log10(abs(src.brightness())) - math.log10(min_bright) + 1)*min_size;
        else:
          ysize = min_size;
        xsize = ysize/(math.cos(src.pos.dec) or 1);
        # figure out source style
        style,label = self.model.getSourcePlotStyle(src);
        if style:
          f.write('# %s\n'%src.name);
          # write symbol for source
          f.write('COLOR %s\n'%style.symbol_color);
          if style.symbol == "plus":
            f.write('CROSS %.12f %.12f %f %f\n'%(ra,dec,xsize,ysize));
          elif style.symbol == "cross":
            f.write('CROSS %.12f %.12f %f %f 45\n'%(ra,dec,ysize,ysize));
          elif style.symbol == "circle":
            f.write('CIRCLE %.12f %.12f %f\n'%(ra,dec,ysize));
          elif style.symbol == "dot":
            f.write('DOT %.12f %.12f\n'%(ra,dec));
          elif style.symbol == "square":
            f.write('CBOX %.12f %.12f %f %f\n'%(ra,dec,xsize,ysize));
          elif style.symbol == "diamond":
            f.write('CBOX %.12f %.12f %f %f 45\n'%(ra,dec,xsize,ysize));
          # write label
          if label:
            f.write('FONT hershey%d\n'%(style.label_size*2));
            f.write('COLOR %s\n'%style.label_color);
            f.write('TEXT %.12f %.12f %s\n'%(ra,dec,label));
      f.close();
    except IOError,err:
      busy = None;
      self.qerrmsg.showMessage("Error writing Karma annotations file %s: %s"%(filename,str(err)));
      return;
    busy = None;
    self.parent().showMessage("Wrote Karma annotations for %d sources to file %s"%(len(sources),filename));
    return QDialog.accept(self);

def export_karma_annotations (mainwin,model):
  dialog = getattr(mainwin,'_export_karma_dialog',None);
  if not dialog:
    dialog = mainwin._export_karma_dialog = ExportKarmaDialog(mainwin);
  dialog.setModel(model);
  # show dialog
  return dialog.exec_();

from TigGUI.Tools import registerTool
registerTool("Export Karma annotations...",export_karma_annotations);


# -*- coding: utf-8 -*-
#
#% $Id$ 
#
#
# Copyright (C) 2002-2011
# The MeqTree Foundation & 
# ASTRON (Netherlands Foundation for Research in Astronomy)
# P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>,
# or write to the Free Software Foundation, Inc., 
# 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

from PyQt4.Qt import *
from PyQt4.Qwt5 import *

import math
import  numpy
import numpy.ma
from scipy.ndimage import measurements

import Kittens.utils
import copy

_verbosity = Kittens.utils.verbosity(name="colormap");
dprint = _verbosity.dprint;
dprintf = _verbosity.dprintf;

class IntensityMap (object):
  """An IntensityMap maps a float array into a 0...1 range."""
  def __init__ (self,dmin=None,dmax=None):
    """Constructor. An optional data range may be supplied.""";
    self.range = None;
    if dmin is not None:
      if dmax is None:
        raise TypeError,"both dmin and dmax must be specified, or neither.""";
      self.setDataRange(dmin,dmax);

  def copy (self):
    return copy.copy(self);

  def setDataRange (self,dmin,dmax):
    """Sets the data range.""";
    self.range = dmin,dmax;

  def setDataSubset (self,subset,minmax=None):
    """Sets the data subset.""";
    self.subset = subset;
    self.subset_minmax = minmax;

  def getDataSubset (self):
    return self.subset,self.subset_minmax;

  def getDataRange (self,data):
    """Returns the set data range, or uses data min/max if it is not set""";
    # use data min/max if no explicit ranges are set
    return self.range or measurements.extrema(data)[:2];

  def remap (self,data):
    """Remaps data into 0...1 range""";
    raise RuntimeError,"remap() not implemented in "+str(type(self));

class LinearIntensityMap (IntensityMap):
  """This scales data linearly between preset min and max values."""
  def remap (self,data):
    d0,d1 = self.getDataRange(data);
    dd = d1 - d0;
    if dd:
      return ((data-d0)/dd).clip(0,1);
    else:
      return numpy.zeros(data.shape,float);

class LogIntensityMap (IntensityMap):
  """This scales data linearly between preset min and max values."""
  def __init__ (self,log_cycles=6):
    self.log_cycles = log_cycles;

  def remap (self,data):
    # d0,d1 is current data range
    d0,d1 = self.getDataRange(data);
    if d0 == d1:
      return numpy.zeros(data.shape,float);
    dmax  = d1 - d0;
    data = data - d0;
    dmin = dmax*(10**(-self.log_cycles));
    # clip data to between dmin and dmax, and take log
    data = numpy.ma.log10(data.clip(dmin,dmax));
    # now rescale
    return (data - math.log10(dmin))/(math.log10(dmax) - math.log10(dmin));


class HistEqIntensityMap (IntensityMap):
  def __init__ (self,nbins=256):
    """Creates intensity mapper which uses histogram equalization.""";
    IntensityMap.__init__(self);
    self._nbins = nbins;
    self._cdf = self._bins = self.subset = None;

  def setDataSubset (self,subset,minmax=None):
    IntensityMap.setDataSubset(self,subset,minmax);
    self._bins = None;  # to recompute the CDF

  def setDataRange (self,*range):
    IntensityMap.setDataRange(self,*range);
    self._bins = None;  # to recompute the CDF

  def _computeCDF (self,data):
    """Recomputes the CDF using the current data subset and range""";
    dmin,dmax = self.getDataRange(self.subset if self.subset is not None else data);
    if dmin == dmax:
      self._cdf = None;
    else:
      dprint(1,"computing CDF for range",dmin,dmax);
      # make cumulative histogram, normalize to 0...1
      hist = measurements.histogram(self.subset if self.subset is not None else data,dmin,dmax,self._nbins);
      cdf = numpy.cumsum(hist);
      cdf = cdf/float(cdf[-1]);
      # append 0 at beginning, as left side of bin
      self._cdf = numpy.zeros(len(cdf)+1,float);
      self._cdf[1:] = cdf[...];
      # make array of bin edges
      self._bins = dmin + (dmax-dmin)*numpy.arange(self._nbins+1)/float(self._nbins);

  def remap (self,data):
    if self._bins is None:
      self._computeCDF(data);
    if self._cdf is None:
      return numpy.zeros(data.shape,float);
    values = numpy.interp(data.ravel(),self._bins,self._cdf).reshape(data.shape);
    if hasattr(data,'mask'):
      values = numpy.ma.masked_array(values,data.mask);
    return values;

class Colormap (QObject):
  """A Colormap provides operations for turning normalized float arrays into QImages. The default implementation is a linear colormap between two colors.
  """;
  def __init__ (self,name,color0=QColor("black"),color1=QColor("white"),alpha=(1,1)):
    QObject.__init__(self);
    self.name = name;
    # color is either specified as one argument (which should then be a [3,n] or [4,n] array),
    # or as two QColors orstring names.
    if isinstance(color0,(list,tuple)):
      self._rgb = numpy.array(color0);
      if self._rgb.shape[1] != 3 or self._rgb.shape[0] < 2:
        raise TypeError,"expected [N,3] (N>=2) array as first argument";
    else:
      if isinstance(color0,str):
        color0 = QColor(color0);
      if isinstance(color1,str):
        color1 = QColor(color1);
      self._rgb = numpy.array([[color0.red(),color0.green(),color0.blue()],
                                             [color1.red(),color1.green(),color1.blue()]])/255.;
    self._rgb_arg = numpy.arange(self._rgb.shape[0])/(self._rgb.shape[0]-1.0)
    # alpha array
    self._alpha = numpy.array(alpha).astype(float);
    self._alpha_arg = numpy.arange(len(alpha))/(len(alpha)-1.0);
    # background brush
    self._brush = None;

  def makeQImage (self,width,height):
    data = numpy.zeros((width,height),float);
    data[...] = (numpy.arange(width)/(width-1.))[:,numpy.newaxis];
    # make brush image -- diag background, with colormap on top
    img = QImage(width,height,QImage.Format_RGB32);
    painter = QPainter(img);
    painter.fillRect(0,0,width,height,QBrush(QColor("white")));
    painter.fillRect(0,0,width,height,QBrush(Qt.BDiagPattern));
    painter.drawImage(0,0,self.colorize(data));
    painter.end();
    return img;

  def makeQPixmap (self,width,height):
    data = numpy.zeros((width,height),float);
    data[...] = (numpy.arange(width)/(width-1.))[:,numpy.newaxis];
    # make brush image -- diag background, with colormap on top
    img = QPixmap(width,height);
    painter = QPainter(img);
    painter.fillRect(0,0,width,height,QBrush(QColor("white")));
    painter.fillRect(0,0,width,height,QBrush(Qt.BDiagPattern));
    painter.drawImage(0,0,self.colorize(data));
    painter.end();
    return img;

  def makeBrush (self,width,height):
    return QBrush(self.makeQImage(width,height));

  def colorize (self,data,alpha=None):
    """Converts normalized data (0...1) array into a QImage of the same dimensions.
    'alpha', if set, is a 0...1 array of the same size, which is mapped to the alpha channel
    (i.e. 0 for fully transparent and 1 for fully opaque).
    If data is a masked array, masked pixels will be fully transparent.""";
    # setup alpha channel
    if alpha is None:
      alpha = numpy.interp(data.ravel(),self._alpha_arg,self._alpha).reshape(data.shape);
    alpha = numpy.round(255*alpha).astype(numpy.int32).clip(0,255);
    # make RGB arrays
    rgbs = [ (numpy.interp(data.ravel(),self._rgb_arg,self._rgb[:,i]).
                reshape(data.shape)*255).round().astype(numpy.int32).clip(0,255)
             for i in range(3) ];
    # add data mask
    mask = getattr(data,'mask',None);
    if mask is not None and mask is not False:
      alpha[mask] = 0;
      for x in rgbs:
        x[mask] = 0;
    # do the deed
    return self.QARGBImage(alpha,*rgbs);
    
  def makeControlWidgets (self,parent):
    """Creates control widgets for the colormap's internal parameters.
    "parent" is a parent widget.
    Returns None if no controls are required""";
    return None;

  class QARGBImage (QImage):
    """This is a QImage which is constructed from an A,R,G,B arrays.""";
    def __init__ (self,a,r,g,b):
      nx,ny = r.shape;
      argb = (a<<24) | (r<<16) | (g<<8) | b;
      # transpose array, as it is in column-major (C order), while QImages are in row-major order
      dprint(5,"making qimage of size",nx,ny);
      self._buffer = argb.transpose().tostring();
      QImage.__init__(self,self._buffer,nx,ny,QImage.Format_ARGB32);

class ColormapWithControls (Colormap):
  """This is a base class for a colormap with controls knobs""";
  class SliderControl (QObject):
    """This class implements a slider control for a colormap""";
    def __init__ (self,name,value,minval,maxval,step,format="%s: %.1f"):
      QObject.__init__(self);
      self.name,self.value,self.minval,self.maxval,self.step,self.format = \
        name,value,minval,maxval,step,format;
      self._default = value;
      self._wlabel = None;

    def makeControlWidgets (self,parent,gridlayout,row,column):
      toprow = QWidget(parent);
      gridlayout.addWidget(toprow,row*2,column);
      top_lo = QHBoxLayout(toprow);
      top_lo.setContentsMargins(0,0,0,0);
      self._wlabel = QLabel(self.format%(self.name,self.value),toprow);
      top_lo.addWidget(self._wlabel);
      self._wreset = QToolButton(toprow);
      self._wreset.setText("reset");
      self._wreset.setToolButtonStyle(Qt.ToolButtonTextOnly);
      self._wreset.setAutoRaise(True);
      self._wreset.setEnabled(self.value != self._default);
      QObject.connect(self._wreset,SIGNAL("clicked()"),self._resetValue);
      top_lo.addWidget(self._wreset);
      self._wslider = QwtSlider(parent);
      # This works around a stupid bug in QwtSliders -- see comments on histogram zoom wheel above
      self._wslider_timer = QTimer(parent);
      self._wslider_timer.setSingleShot(True);
      self._wslider_timer.setInterval(500);
      QObject.connect(self._wslider_timer,SIGNAL("timeout()"),self.setValue);
      gridlayout.addWidget(self._wslider,row*2+1,column);
      self._wslider.setRange(self.minval,self.maxval);
      self._wslider.setStep(self.step);
      self._wslider.setValue(self.value);
      self._wslider.setTracking(False);
      QObject.connect(self._wslider,SIGNAL("valueChanged(double)"),self.setValue);
      QObject.connect(self._wslider,SIGNAL("sliderMoved(double)"),self._previewValue);

    def _resetValue (self):
      self._wslider.setValue(self._default);
      self.setValue(self._default);

    def setValue (self,value=None,notify=True):
      # only update widgets if already created
      self.value = value;
      if self._wlabel is not None:
        if value is None:
          self.value = value = self._wslider.value();
        self._wreset.setEnabled(value != self._default);
        self._wlabel.setText(self.format%(self.name,self.value));
        # stop timer if being called to finalize the change in value
        if notify:
          self._wslider_timer.stop();
          self.emit(SIGNAL("valueChanged"),self.value);
      
    def _previewValue (self,value):
      self.setValue(notify=False);
      self._wslider_timer.start(500);
      self.emit(SIGNAL("valueMoved"),self.value);
      
  def emitChange (self,*dum):
    self.emit(SIGNAL("colormapChanged"));

  def emitPreview (self,*dum):
    self.emit(SIGNAL("colormapPreviewed"));
    
  def loadConfig (self,config):
    pass;
    
  def saveConfig (self,config,save=True):
    pass;

class CubeHelixColormap (ColormapWithControls):
  """This implements the "cubehelix" colour scheme proposed by Dave Green:
  D. Green 2011, Bull. Astr. Soc. India (2011) 39, 289–295
  http://arxiv.org/pdf/1108.5083v1
  """
  def __init__(self,gamma=1,rgb=0.5,rots=-1.5,hue=1.2,name="CubeHelix"):
    ColormapWithControls.__init__(self,name);
    self.gamma  = self.SliderControl("Gamma",gamma,0,6,.1);
    self.color  = self.SliderControl("Colour",rgb,0,3,.1);
    self.cycles = self.SliderControl("Cycles",rots,-10,10,.1);
    self.hue    = self.SliderControl("Hue",hue,0,2,.1);
    
  def colorize (self,data,alpha=None):
    """Converts normalized data (0...1) array into a QImage of the same dimensions.
    'alpha', if set, is a 0...1 array of the same size, which is mapped to the alpha channel
    (i.e. 0 for fully transparent and 1 for fully opaque).
    If data is a masked array, masked pixels will be fully transparent.""";
    # setup alpha channel
    if alpha is None:
      alpha = numpy.zeros(data.shape,dtype=numpy.int32);
      alpha[...] = 255;
    else:
      alpha = numpy.round(255*alpha).astype(numpy.int32).clip(0,255);
    # make RGB arrays
    dg = data**self.gamma.value;
    a = self.hue.value*dg*(1-dg)/2;
    phi = 2*math.pi*(self.color.value/3 + self.cycles.value*data);
    cosphi = a*numpy.cos(phi);
    sinphi = a*numpy.sin(phi);
    r = dg - 0.14861*cosphi + 1.78277*sinphi;
    g = dg - 0.29227*cosphi - 0.90649*sinphi;
    b = dg + 1.97249*cosphi;
    rgbs = [ (x*255).round().astype(numpy.int32).clip(0,255) for x in r,g,b ];
    # add data mask
    mask = getattr(data,'mask',None);
    if mask is not None and mask is not False:
      alpha[mask] = 0;
      for x in rgbs:
        x[mask] = 0;
    # do the deed
    return self.QARGBImage(alpha,*rgbs);

  def makeControlWidgets (self,parent):
    """Creates control widgets for the colormap's internal parameters.
    "parent" is a parent widget.
    Returns None if no controls are required""";
    top = QWidget(parent);
    layout = QGridLayout(top);
    layout.setContentsMargins(0,0,0,0);
    for irow,icol,control in ((0,0,self.gamma),(0,1,self.color),(1,0,self.cycles),(1,1,self.hue)):
      control.makeControlWidgets(top,layout,irow,icol);
      QObject.connect(control,SIGNAL("valueChanged"),self.emitChange);
      QObject.connect(control,SIGNAL("valueMoved"),self.emitPreview);
    return top;

  def loadConfig (self,config):
    for name in "gamma","color","cycles","hue":
      control = getattr(self,name);
      value = config.getfloat("cubehelix-colourmap-%s"%name,control.value);
      control.setValue(value,notify=False);
    
  def saveConfig (self,config,save=True):
    for name in "gamma","color","cycles","hue":
      control = getattr(self,name);
      config.set("cubehelix-colourmap-%s"%name,control.value,save=save);


# instantiate "static" colormaps (i.e. those that have no internal parameters, and thus can be
# shared among images without instantiating a new Colormap object for each)
GreyscaleColormap = Colormap("Greyscale");
TransparentFuchsiaColormap = Colormap("Transparent Fuchsia",color0="fuchsia",color1="fuchsia",alpha=(0,1));

from ColormapTables import Karma
_karma_colormaps = [
    Colormap(cmap,getattr(Karma,cmap))
      for cmap in [
        "Background",
        "Heat",
        "Isophot",
        "Mousse",
        "Rainbow",
        "RGB",
        "RGB2",
        "Smooth",
        "Staircase",
        "Mirp",
        "Random" ]
  ];
  
def getColormapList ():
  """Returns list of Colormap instances."""
  
  # Some colormaps need a unique instantiation (because they have parameters)
  # For the rest, use the static objects
  return [ GreyscaleColormap,
           CubeHelixColormap(),
           TransparentFuchsiaColormap ] + _karma_colormaps;
